
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rti_me_c.h"
#include "wh_sm/wh_sm_history.h"
#include "rh_sm/rh_sm_history.h"

#include "ShapeType.h"
#include "ShapeTypeSupport.h"
#include "ShapeTypePlugin.h"
#include "ShapeTypeApplication.h"

/* See Wire Protocol Specification on http://www.omg.org/spec/DDSI-RTPS/2.2/
* for more details about CDR encapsulation.
*/

/*ci \brief OMG CDR Big Endian encapsulation ID */
#define ENCAPSULATION_ID_CDR_BE     0x0000

/*ci \brief OMG CDR Little Endian encapsulation ID */
#define ENCAPSULATION_ID_CDR_LE     0x0001

/*ci \brief Empty encapsulation options */
#define ENCAPSULATION_OPTIONS_NONE  0x0000

/*ci \brief Unsigned Short type size */
#define UNSIGNED_SHORT_SIZE 2

/*ci \brief Unsigned Long type size */
#define UNSIGNED_LONG_SIZE  4


// We will use this values to filter the samples (-1 indicates that no filter is applied)
CDR_Long xMin = -1; 
CDR_Long xMax = -1;
CDR_Long yMin = -1;
CDR_Long yMax = -1;

// Colors that must be dropped
char **sColor;
int nColors = 0;



void
ShapeTypeExtendedSubscriber_on_data_available(void *listener_data,
DDS_DataReader * reader)
{
    ShapeTypeExtendedDataReader *hw_reader = ShapeTypeExtendedDataReader_narrow(reader);
    DDS_ReturnCode_t retcode;
    struct DDS_SampleInfo *sample_info = NULL;
    ShapeTypeExtended *sample = NULL;

    struct DDS_SampleInfoSeq info_seq = 
    DDS_SEQUENCE_INITIALIZER;
    struct ShapeTypeExtendedSeq sample_seq = 
    DDS_SEQUENCE_INITIALIZER;

    const DDS_Long TAKE_MAX_SAMPLES = 32;
    DDS_Long i;

    retcode = ShapeTypeExtendedDataReader_take(hw_reader, 
    &sample_seq, &info_seq, TAKE_MAX_SAMPLES, 
    DDS_ANY_SAMPLE_STATE, DDS_ANY_VIEW_STATE, DDS_ANY_INSTANCE_STATE);

    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed to take data, retcode(%d)\n", retcode);
        goto done;
    }

    /* Print each valid sample taken */
    for (i = 0; i < ShapeTypeExtendedSeq_get_length(&sample_seq); ++i)
    {
        sample_info = DDS_SampleInfoSeq_get_reference(&info_seq, i);

        if (sample_info->valid_data)
        {
            sample = ShapeTypeExtendedSeq_get_reference(&sample_seq, i);
			
			// Print the data sample
			printf("Received square: Color: %s, Position: (x=%d, y=%d)\n", ((ShapeType*) sample)->color, ((ShapeType*) sample)->x, ((ShapeType*) sample)->y);
        }
        else
        {
            printf("\nSample received\n\tINVALID DATA\n");
        }
    }

    ShapeTypeExtendedDataReader_return_loan(hw_reader, &sample_seq, &info_seq);

    done:
    ShapeTypeExtendedSeq_finalize(&sample_seq);
    DDS_SampleInfoSeq_finalize(&info_seq);
}

void
ShapeTypeExtendedSubscriber_on_subscription_matched(void *listener_data,
DDS_DataReader * reader,
const struct
DDS_SubscriptionMatchedStatus
*status)
{
    if (status->current_count_change > 0)
    {
        printf("Matched a publisher\n");
    }
    else if (status->current_count_change < 0)
    {
        printf("Unmatched a publisher\n");
    }
}

/*i
* \brief Helper function to filter an ShapeTypeExtended sample
*
* \param[in]  sample       A ShapeTypeExtended data sample to filter
* \param[out] drop_sample  Out parameter determining whether the sample
*                          should be filtered out or not.
*/
void 
ShapeTypeExtendedSubscriber_filter_sample(ShapeTypeExtended *sample,
DDS_Boolean *drop_sample)
{
    #if 0
    /* Example filter: drop samples with even-numbered count in id */

    *drop_sample = (sample->id % 2 == 0) ? DDS_BOOLEAN_TRUE : DDS_BOOLEAN_FALSE;

    #else

    *drop_sample = DDS_BOOLEAN_FALSE;

    #endif
}

/*i
* \brief Helper function to deserialize an unsigned short
*
* \param[in]  src_buffer      Pointer to CDR stream buffer
* \param[in]  need_byte_swap  Indicates whether it is needed to swap bytes
* \param[out] instance        Deserialized unsigned short
*/
void
ShapeTypeExtendedSubscriber_deserialize_unsigned_short(char **src_buffer,
RTI_BOOL need_byte_swap,
RTI_UINT16 *instance)
{
    if (need_byte_swap)
    {
        *((RTI_INT8*)instance + 1) = *((*src_buffer)++);
        *((RTI_INT8*)instance    ) = *((*src_buffer)++);
    }
    else
    {
        *instance = *(RTI_UINT16*)(*src_buffer);
        (*src_buffer) += CDR_SHORT_SIZE;
    }
}

/*i
* \brief Helper function to deserialize an unsigned short in big endian from
*        a CDR stream buffer
*
* \param[in]  src_buffer      Pointer to CDR stream buffer
* \param[out] instance        Deserialized unsigned short
*/
void
ShapeTypeExtendedSubscriber_deserialize_unsigned_short_from_big_endian(
    char **src_buffer,
    RTI_UINT16 *instance)
{
    #ifdef RTI_ENDIAN_LITTLE
    /* LE: always swap */
    ShapeTypeExtendedSubscriber_deserialize_unsigned_short(src_buffer,
    RTI_TRUE,
    instance);
    #else
    /* BE: never swap */
    ShapeTypeExtendedSubscriber_deserialize_unsigned_short(src_buffer,
    RTI_FALSE,
    instance);
    #endif
}

/*i
* \brief Helper function to deserialize an unsigned long
*
* \param[in]  src_buffer      Pointer to CDR stream buffer
* \param[in]  need_byte_swap  Indicates whether it is needed to swap bytes
* \param[out] instance        Deserialized unsigned long
*/
void
ShapeTypeExtendedSubscriber_deserialize_unsigned_long(char **src_buffer,
RTI_BOOL need_byte_swap,
DDS_UnsignedLong *instance)
{
    RTI_INT32 i;
    if (need_byte_swap)
    {
        for (i = 3; i >= 0; --i)
        {
            *((RTI_INT8*)instance + i) = *((*src_buffer)++);
        }
    }
    else
    {
        *instance = *(RTI_UINT32*)(*src_buffer);
        (*src_buffer) += CDR_LONG_SIZE;
    }
}

/*i
* \brief Helper function to deserialize the CDR stream header
*
* \param[in]      src_buffer      Pointer to CDR stream buffer
* \param[in/out]  endian          CDR stream endianness
* \param[in/out]  need_byte_swap  Indicates whether it is needed to swap bytes
*
* \return RTI_TRUE if the header is correctly deserialized, otherwise RTI_FALSE
*/
DDS_Boolean
ShapeTypeExtendedSubscriber_deserialize_header(char **src_buffer,
RTI_UINT8 *endian,
RTI_BOOL *need_byte_swap)
{
    RTI_UINT16 kind;
    RTI_UINT16 options;
    RTI_BOOL   ret_value;

    ret_value = RTI_FALSE;

    ShapeTypeExtendedSubscriber_deserialize_unsigned_short_from_big_endian(
        src_buffer, &kind);

    /* It is not defined whether the options are byte-swapped or not. It is not
    * important at the moment as the only supported value is 0.
    */
    ShapeTypeExtendedSubscriber_deserialize_unsigned_short(src_buffer,
    *need_byte_swap,
    &options);

    /* No CDR options are supported */
    if (options == ENCAPSULATION_OPTIONS_NONE)
    {
        switch (kind)
        {
            case ENCAPSULATION_ID_CDR_BE:
            if ((*endian) != RTI_CDR_ENDIAN_BIG)
            {
                *endian = RTI_CDR_ENDIAN_BIG;
                *need_byte_swap =
                (*need_byte_swap == RTI_TRUE) ? RTI_FALSE : RTI_TRUE;
            }
            ret_value = RTI_TRUE;
            break;
            case ENCAPSULATION_ID_CDR_LE:
            if ((*endian) != RTI_CDR_ENDIAN_LITTLE)
            {
                *endian = RTI_CDR_ENDIAN_LITTLE;
                *need_byte_swap =
                (*need_byte_swap == RTI_TRUE) ? RTI_FALSE : RTI_TRUE;
            }
            ret_value = RTI_TRUE;
            break;
            default:
            break;
        }
    }

    return ret_value;
}


/*i
 *\brief Helper function to check if the sample must be rejected or not before deserialize
 *
 *\param[in]	src_buffer		Pointer to CDR stream buffer
 *\param[in]	need_byte_swap	Indicates wheter it is need to swap bytes
 *\return		DDS_Boolean		Boolean indicating if the sample should be dropped (true) or not (false)
 */
DDS_Boolean
ShapeTypeSubscriber_check_String(char **src_buffer,
		                                           RTI_BOOL need_byte_swap){
    
	int i = 0;

	// To get the color from the sample, the first thing that we must do is read the length of the string that defines the color
	RTI_UINT32 length;
	ShapeTypeExtendedSubscriber_deserialize_unsigned_long(src_buffer, need_byte_swap, &length);

	// and compare the string in the sample with the color that we are dropping
	
	for(i=0; i<nColors; i++){
		if(strncmp((*src_buffer), sColor[i], length) == 0){
			return RTI_TRUE;
		}
	}
	return RTI_FALSE;	
}


/*
 * FIRST FILTER - on_before_sample_deserialize
 *
 * Filtering the sample before deserialize it - we will reject the sample depending on the color of the shape (the user must indicate the color).
 * If the criteria to reject the sample is meeted, we must set the sample dropped boolean to true. The sample will be dropped after the execution of the callback before deserialize
 * If the criteria to reject the sample is not meeted, we must set the sample dropped boolean to false. The sample will be available to deserialization and on_before_sample_commit will be executed before read/take
 */
DDS_Boolean shapes_on_before_sample_deserialize(void *listener_data, DDS_DataReader *reader, struct NDDS_Type_Plugin *plugin, struct CDR_Stream_t *stream, DDS_Boolean *sample_dropped){
	// As we are going to filter depending only on one component of the sample, we must deserialize only that component of the sample
	DDS_Long color = 0;
	RTI_BOOL need_byte_swap = DDS_BOOLEAN_FALSE;
	RTI_UINT16 endian = 0;
	char *src_buffer = NULL;
	
	need_byte_swap = stream->need_byte_swap;
	endian=stream->endian;
	src_buffer = CDR_Stream_get_current_position_ptr(stream);

	// Deserialize the sample header
	if(!ShapeTypeExtendedSubscriber_deserialize_header(&src_buffer, &endian, &need_byte_swap)){
		return DDS_BOOLEAN_FALSE;
	}

	// Check if we must reject or not the sample
	*sample_dropped = ShapeTypeSubscriber_check_String(&src_buffer, need_byte_swap);

	return DDS_BOOLEAN_TRUE;
}


/*
 * SECOND FILTER - on_before_sample_commit
 *
 * Filetring the data - we will reject samples depending on the values X and Y of the shapes
 * If the rejecting criteria is meeted, set the sample dropped boolean to true. This will drop the sample after the execution of the callback
 * To keep the sample we must set the sample dropped boolean to false, this way we will keep the sample and we will be able to read() or take() it later
 */
DDS_Boolean shapes_on_before_sample_commit(void *listener_data, DDS_DataReader *reader, const void *const sample, const struct DDS_SampleInfo *const sample_info, DDS_Boolean *sample_dropped)
{
	
	ShapeType *actualShape = (ShapeType *) sample;

	if(xMin != -1 && actualShape->x < xMin){
		// x < xMin
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else if(xMax != -1 && actualShape->x > xMax){
		// x > xMax
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else if(yMin != -1 && actualShape->y < yMin){
		// y < yMin
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else if(yMax != -1 && actualShape->y > yMax){
		// y > yMax
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else{	
		// Sample should not be dropped
		*sample_dropped = DDS_BOOLEAN_FALSE;
	}
	
	// Print a message indicating that the sample has been dropped
	/*if(*sample_dropped){
		printf("on_before_sample_commit: Dropped sample - position (x=%d, y=%d)\n", actualShape->x, actualShape->y);
	}*/
	return DDS_BOOLEAN_TRUE;
}



int
subscriber_main_w_args(DDS_Long domain_id, char *udp_intf, char *peer,
DDS_Long sleep_time, DDS_Long count)
{
    DDS_Subscriber *subscriber;
    DDS_DataReader *datareader;
    struct DDS_DataReaderQos dr_qos = DDS_DataReaderQos_INITIALIZER;
    DDS_ReturnCode_t retcode;
    struct DDS_DataReaderListener dr_listener =
    DDS_DataReaderListener_INITIALIZER;
    struct Application *application;

    application = Application_create("subscriber", "publisher", domain_id,
    udp_intf, peer, sleep_time, count);

    if (application == NULL)
    {
        return 0;
    }

    subscriber =
    DDS_DomainParticipant_create_subscriber(application->participant,
    &DDS_SUBSCRIBER_QOS_DEFAULT,
    NULL, DDS_STATUS_MASK_NONE);
    if (subscriber == NULL)
    {
        printf("subscriber == NULL\n");
        goto done;
    }

    /* Uncomment to filter samples with callbacks */
    /* #define HW_USE_SAMPLE_FILTER */
    #ifdef HW_USE_SAMPLE_FILTER
    /* choose one callback to enable */
    #if 1
    dr_listener.on_before_sample_deserialize = 
    ShapeTypeExtendedSubscriber_on_before_sample_deserialize;
    #else
    dr_listener.on_before_sample_commit = 
    ShapeTypeExtendedSubscriber_on_before_sample_commit;
    #endif

    #endif

    /* INTEROPERABILITY NOTE:
    Non-Connext Micro DDS writers and readers will have default LivelinessQoS
    kind set to DDS_AUTOMATIC_LIVELINESS_QOS.

    Because Connext Micro currently only supports
    DDS_MANUAL_BY_TOPIC_LIVELINESS_QOS for finite lease_duration,
    a non-Connext Micro writer must set its liveliness kind to
    DDS_MANUAL_BY_TOPIC_LIVELINESS_QOS as well in order for
    Requested-Offered (RxO) semantics to be compatible.
    */

    dr_listener.on_data_available = ShapeTypeExtendedSubscriber_on_data_available;
    dr_listener.on_subscription_matched =
    ShapeTypeExtendedSubscriber_on_subscription_matched;


	/************ Assign the callbacks to the datareader listener ************/
	// Assigning the callback for on_before_sample_commit
	dr_listener.on_before_sample_commit = shapes_on_before_sample_commit;
	// Asigning the callback for the on_before_sample_deserialize
	dr_listener.on_before_sample_deserialize = shapes_on_before_sample_deserialize;


    #ifdef HW_USE_SAMPLE_FILTER
    dr_qos.resource_limits.max_instances = 1;
    #else
    dr_qos.resource_limits.max_instances = 2;
    #endif

    dr_qos.resource_limits.max_samples_per_instance = 32;
    dr_qos.resource_limits.max_samples = dr_qos.resource_limits.max_instances *
    dr_qos.resource_limits.max_samples_per_instance;
    /* if there are more remote writers, you need to increase these limits */
    dr_qos.reader_resource_limits.max_remote_writers = 10;
    dr_qos.reader_resource_limits.max_remote_writers_per_instance = 10;
    dr_qos.history.depth = 32;

    /* Reliability QoS */
    #if 1
    dr_qos.reliability.kind = DDS_RELIABLE_RELIABILITY_QOS;
    #else
    dr_qos.reliability.kind = DDS_BEST_EFFORT_RELIABILITY_QOS;
    #endif

    datareader = DDS_Subscriber_create_datareader(subscriber,
    DDS_Topic_as_topicdescription
    (application->topic), &dr_qos,
    &dr_listener,
    DDS_DATA_AVAILABLE_STATUS |
    DDS_SUBSCRIPTION_MATCHED_STATUS);

    if (datareader == NULL)
    {
        printf("datareader == NULL\n");
        goto done;
    }

    retcode = Application_enable(application);
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed to enable application\n");
        goto done;
    }

    if (application->count != 0)
    {
        printf("Running for %d seconds, press Ctrl-C to exit\n",
        application->count);
        OSAPI_Thread_sleep(application->count * 1000);
    }
    else
    {
        int sleep_loop_count =  (24 * 60 * 60) / 2000;
        int sleep_loop_left = (24 * 60 * 60) % 2000;

        printf("Running for 24 hours, press Ctrl-C to exit\n");

        while (sleep_loop_count)
        {
            OSAPI_Thread_sleep(2000  * 1000); /* sleep is in ms */
            --sleep_loop_count;
        }

        OSAPI_Thread_sleep(sleep_loop_left * 1000);
    }

    done:

    Application_delete(application);

    DDS_DataReaderQos_finalize(&dr_qos);

    return 0;
}

#if !(defined(RTI_VXWORKS) && !defined(__RTP__))
int
main(int argc, char **argv)
{
    DDS_Long i = 0;
    DDS_Long domain_id = 0;
    char *peer = NULL;
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;
	int aux_i = 0;

	sColor = (char **) malloc(20 * sizeof(char));
    for (i = 1; i < argc; ++i)
    {
        if (!strcmp(argv[i], "-domain"))
        {
            ++i;
            if (i == argc)
            {
                printf("-domain <domain_id>\n");
                return -1;
            }
            domain_id = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-udp_intf"))
        {
            ++i;
            if (i == argc)
            {
                printf("-udp_intf <interface>\n");
                return -1;
            }
            udp_intf = argv[i];
        }
        else if (!strcmp(argv[i], "-peer"))
        {
            ++i;
            if (i == argc)
            {
                printf("-peer <address>\n");
                return -1;
            }
            peer = argv[i];
        }
        else if (!strcmp(argv[i], "-sleep"))
        {
            ++i;
            if (i == argc)
            {
                printf("-sleep_time <sleep_time>\n");
                return -1;
            }
            sleep_time = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-count"))
        {
            ++i;
            if (i == argc)
            {
                printf("-count <count>\n");
                return -1;
            }
            count = strtol(argv[i], NULL, 0);
        }
		else if(!strcmp(argv[i], "-rejectColor"))
		{
			++i;
			if(i == argc)
			{
				printf("-rejectColor <Color_to_reject>");
				return -1;
			}
			sColor[nColors]= (char *) malloc(20 * sizeof(char));
			strcpy(sColor[nColors], argv[i]);
			aux_i = 0;
			while(sColor[nColors][aux_i]){
				sColor[nColors][aux_i] = toupper(sColor[nColors][aux_i]);
				aux_i++;
			}
			printf("Filter criteria: Color must not be: %s\n", sColor[nColors]);
			nColors++;
		}
		else if(!strcmp(argv[i], "-xMin")){
			++i;
			if(i == argc)
			{
				printf("-xMin <X minumum value>");
				return -1;
			}
			xMin = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Minimum x value: %d\n", xMin);
		}
		else if(!strcmp(argv[i], "-yMin")){
			++i;
			if(i == argc)
			{
				printf("-yMin <Y minumum value>");
				return -1;
			}
			yMin = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Minimum y value: %d\n", yMin);
		}
		else if(!strcmp(argv[i], "-xMax")){
			++i;
			if(i == argc)
			{
				printf("-xMax <X maximum value>");
				return -1;
			}
			xMax = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Maximum X value: %d\n", xMax);
		}
		else if(!strcmp(argv[i], "-yMax")){
			++i;
			if(i == argc)
			{
				printf("-yMax <Y maximum value>");
				return -1;
			}
			yMax = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Maximum Y value: %d\n", yMax);
		}
        else if (!strcmp(argv[i], "-h"))
        {
            Application_help(argv[0]);
            return 0;
        }
        else
        {
            printf("unknown option: %s\n", argv[i]);
            return -1;
        }
    }

    return subscriber_main_w_args(domain_id, udp_intf, peer, sleep_time, count);
}
#elif defined(RTI_VXWORKS)
int
subscriber_main(void)
{
    /* Explicitly configure args below */
    DDS_Long domain_id = 44;
    char *peer = "10.10.65.103";
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;

    return subscriber_main_w_args(domain_id, udp_intf, peer, sleep_time, count);
}
#endif
