
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef rti_me_cpp_hxx
#include "rti_me_cpp.hxx"
#endif
#include "wh_sm/wh_sm_history.h"
#include "rh_sm/rh_sm_history.h"

#include "ShapeType.h"
#include "ShapeTypeSupport.h"
#include "ShapeTypePlugin.h"
#include "ShapeTypeApplication.h"


class ShapeTypeExtendedReaderListener : public DDSDataReaderListener
{

  public:
    virtual void on_data_available(DDSDataReader *reader);

	// on_before_sample_deserialize : This callback is applied after receive the sample and before apply the deserialization. Compared to on_before_sample_commit we will
	// be able to drop the samples with less processing as in the moment this callback is applied the sample is not deserialized. In our example, we will reject samples
	// of certain color before apply the deserialization.
	virtual DDS_Boolean on_before_sample_deserialize(DDSDataReader * reader, NDDS_Type_Plugin * plugin, CDR_Stream_t * stream_t, DDS_Boolean * sample_dropped);


	// on_before_sample_commit : The sample has been already deserialized and we have access to all the data of the sample. We can use this callback to filter using the
	// data available in the sample. In our case, we will check the values for X and Y to drop it or not. This filter is applied before commit the sample to the application
	// logic.
	virtual DDS_Boolean on_before_sample_commit(DDSDataReader * reader,const void *const sample,const struct DDS_SampleInfo *const sample_info,DDS_Boolean * sample_dropped);

};

// We will use this values to filter the samples (-1 indicates that no filter is applied)
CDR_Long xMin = -1; 
CDR_Long xMax = -1;
CDR_Long yMin = -1;
CDR_Long yMax = -1;

// Colors that must be dropped
char **sColor = new char*[20];
int nColors = 0;

/*ci \brief OMG CDR Big Endian encapsulation ID */
#define ENCAPSULATION_ID_CDR_BE     0x0000

/*ci \brief OMG CDR Little Endian encapsulation ID */
#define ENCAPSULATION_ID_CDR_LE     0x0001

/*ci \brief Empty encapsulation options */
#define ENCAPSULATION_OPTIONS_NONE  0x0000

// Samples will be publisher in a second domain, the user can enable this option with the parameter -publishDomain
DDSPublisher *publisher = NULL;


template <typename T>
void take_and_print(typename T::DataReader* reader)
{
    DDS_ReturnCode_t retcode;
    struct DDS_SampleInfo sample_info;
    T *sample = NULL;

    sample = T::TypeSupport::create_data();
    if (sample == NULL)
    {
        printf("Failed sample initialize\n");
        return;
    }

    retcode = reader->take_next_sample(*sample, sample_info);
    while (retcode == DDS_RETCODE_OK)
    {
        if (sample_info.valid_data)
        {
            printf("Received square: Color: %s, Position: (x=%d, y=%d)\n", sample->color, sample->x, sample->y);
			if (retcode != DDS_RETCODE_OK)
			{
				printf("Failed to write sample\n");
			} 
        }
        else
        {
            printf("\nSample received\n\tINVALID DATA\n");
        }
        retcode = reader->take_next_sample(*sample, sample_info);
    }

    T::TypeSupport::delete_data(sample);
}


/*i
 * \brief Helper function to deserialize an unsigned short
 *
 * \param[in]  src_buffer      Pointer to CDR stream buffer
 * \param[in]  need_byte_swap  Indicates whether it is needed to swap bytes
 * \param[out] instance        Deserialized unsigned short
 */
void
ShapeTypeSubscriber_deserialize_unsigned_short(char **src_buffer,
		                                           RTI_BOOL need_byte_swap,
			                                        RTI_UINT16 *instance)
{
    if (need_byte_swap)
    {
        *((RTI_INT8*)instance + 1) = *((*src_buffer)++);
        *((RTI_INT8*)instance    ) = *((*src_buffer)++);
    }
    else
    {
        *instance = *(RTI_UINT16*)(*src_buffer);
        (*src_buffer) += CDR_SHORT_SIZE;
    }
}


/*i
 * \brief Helper function to deserialize an unsigned long
 *
 * \param[in]	src_buffer		Pointer to CDR stream buffer
 * \param[in]	need_byte_swap	Indicates wheter it is need to swap bytes
 * \param[out]	instance		Deserialized unsigned long
 */
void
ShapeTypeSubscriber_deserialize_unsigned_long(char **src_buffer, RTI_BOOL need_byte_swap, RTI_UINT32 *instance){

	if(need_byte_swap){
		*((RTI_INT8*)instance + 3) = *((*src_buffer)++);
		*((RTI_INT8*)instance + 2) = *((*src_buffer)++);
		*((RTI_INT8*)instance + 1) = *((*src_buffer)++);
		*((RTI_INT8*)instance) = *((*src_buffer)++);
	}

	else{
		*instance = *(RTI_UINT32*)(*src_buffer);
		(*src_buffer) += CDR_LONG_SIZE;
	}
}


/*i
 *\brief Helper function to check if the sample must be rejected or not before deserialize
 *
 *\param[in]	src_buffer		Pointer to CDR stream buffer
 *\param[in]	need_byte_swap	Indicates wheter it is need to swap bytes
 *\return		DDS_Boolean		Boolean indicating if the sample should be dropped (true) or not (false)
 */
DDS_Boolean
ShapeTypeSubscriber_check_String(char **src_buffer,
		                                           RTI_BOOL need_byte_swap){
    
	// To get the color from the sample, the first thing that we must do is read the length of the string that defines the color
	RTI_UINT32 length;
	ShapeTypeSubscriber_deserialize_unsigned_long(src_buffer, need_byte_swap, &length);

	// and compare the string in the sample with the color that we are dropping
	for(int i=0; i<nColors; i++){
		if(strncmp((*src_buffer), sColor[i], length) == 0){
			return RTI_TRUE;
		}
	}
	return RTI_FALSE;
	
}


/*i
 * \brief Helper function to deserialize an unsigned short in big endian from
 *        a CDR stream buffer
 *
 * \param[in]  src_buffer      Pointer to CDR stream buffer
 * \param[out] instance        Deserialized unsigned short
 */
void ShapeTypeSubscriber_deserialize_unsigned_short_from_big_endian(
				                                           char **src_buffer,
					                                        RTI_UINT16 *instance)
{
	#ifdef RTI_ENDIAN_LITTLE
		/* LE: always swap */
		ShapeTypeSubscriber_deserialize_unsigned_short(src_buffer,
														RTI_TRUE,
														instance);
	#else
		/* BE: never swap */
		ShapeTypeSubscriber_deserialize_unsigned_short(src_buffer,
														RTI_FALSE,
														instance);
	#endif
}



/*i
 * \brief Helper function to deserialize the CDR stream header
 *
 * \param[in]      src_buffer      Pointer to CDR stream buffer
 * \param[in/out]  endian          CDR stream endianness
 * \param[in/out]  need_byte_swap  Indicates whether it is needed to swap bytes
 *
 * \return RTI_TRUE if the header is correctly deserialized, otherwise RTI_FALSE
 */
DDS_Boolean
ShapeTypeSubscriber_deserialize_header(char **src_buffer,
                                        RTI_UINT16 *endian,
                                        RTI_BOOL *need_byte_swap)
{
    RTI_UINT16 kind;
    RTI_UINT16 options;
    RTI_BOOL   ret_value;

    ret_value = RTI_FALSE;

    ShapeTypeSubscriber_deserialize_unsigned_short_from_big_endian(
                                                            src_buffer, &kind);

    ShapeTypeSubscriber_deserialize_unsigned_short(src_buffer,
                                                    *need_byte_swap,
                                                    &options);

	if (options == ENCAPSULATION_OPTIONS_NONE)
    {
        switch (kind)
        {
            case ENCAPSULATION_ID_CDR_BE:
				if ((*endian) != RTI_CDR_ENDIAN_BIG)
                {
                    *endian = RTI_CDR_ENDIAN_BIG;
                    *need_byte_swap =
                          (*need_byte_swap == RTI_TRUE) ? RTI_FALSE : RTI_TRUE;
                }
                ret_value = RTI_TRUE;
                break;
            case ENCAPSULATION_ID_CDR_LE:
				if ((*endian) != RTI_CDR_ENDIAN_LITTLE)
                {
                    *endian = RTI_CDR_ENDIAN_LITTLE;
                    *need_byte_swap =
                          (*need_byte_swap == RTI_TRUE) ? RTI_FALSE : RTI_TRUE;
                }
                ret_value = RTI_TRUE;
                break;
            default:
                break;
        }
    }

    return ret_value;
}



void
ShapeTypeExtendedReaderListener::on_data_available(DDSDataReader * reader)
{
    ShapeTypeExtendedDataReader *hw_reader = ShapeTypeExtendedDataReader::narrow(reader);
    take_and_print<ShapeTypeExtended>(hw_reader);
}

/*
 * Filtering the sample before deserialize it - we will reject the sample depending on the color of the shape (the user must indicate the color).
 * If the criteria to reject the sample is meeted, we must set the sample dropped boolean to true. The sample will be dropped after the execution of the callback before deserialize
 * If the criteria to reject the sample is not meeted, we must set the sample dropped boolean to false. The sample will be available to deserialization and on_before_sample_commit will be executed before read/take
 */
DDS_Boolean ShapeTypeExtendedReaderListener::on_before_sample_deserialize(DDSDataReader * reader, NDDS_Type_Plugin * plugin, CDR_Stream_t * stream, DDS_Boolean * sample_dropped){
	
	// As we are going to filter depending only on one component of the sample, we must deserialize only that component of the sample
	DDS_Long color = 0;
	RTI_BOOL need_byte_swap = DDS_BOOLEAN_FALSE;
	RTI_UINT16 endian = 0;
	char *src_buffer = NULL;
	
	need_byte_swap = stream->need_byte_swap;
	endian=stream->endian;
	src_buffer = CDR_Stream_get_current_position_ptr(stream);

	// Deserialize the sample header
	if(!ShapeTypeSubscriber_deserialize_header(&src_buffer, &endian, &need_byte_swap)){
		return DDS_BOOLEAN_FALSE;
	}

	// Check if we must reject or not the sample
	*sample_dropped = ShapeTypeSubscriber_check_String(&src_buffer, need_byte_swap);

	return DDS_BOOLEAN_TRUE;
}


/* 
 * Filetring the data - we will reject samples depending on the values X and Y of the shapes
 * If the rejecting criteria is meeted, set the sample dropped boolean to true. This will drop the sample after the execution of the callback
 * To keep the sample we must set the sample dropped boolean to false, this way we will keep the sample and we will be able to read() or take() it later
 */
DDS_Boolean ShapeTypeExtendedReaderListener::on_before_sample_commit(DDSDataReader * reader,const void *const sample,const struct DDS_SampleInfo *const sample_info,DDS_Boolean * sample_dropped){

	
	ShapeType *actualShape = (ShapeType *) sample;

	if(xMin != -1 && actualShape->x < xMin){
		// x < xMin
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else if(xMax != -1 && actualShape->x > xMax){
		// x > xMax
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else if(yMin != -1 && actualShape->y < yMin){
		// y < yMin
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else if(yMax != -1 && actualShape->y > yMax){
		// y > yMax
		*sample_dropped = DDS_BOOLEAN_TRUE;
	}else{	
		// Sample should not be dropped
		*sample_dropped = DDS_BOOLEAN_FALSE;
	}
	
	// Print a message indicating that the sample has been dropped
	/*if(*sample_dropped){
		printf("on_before_sample_commit: Dropped sample - position (x=%d, y=%d)\n", actualShape->x, actualShape->y);
	}*/
	return DDS_BOOLEAN_TRUE;
}


int
subscriber_main_w_args(DDS_Long domain_id, DDS_Long publicationDomain, char *udp_intf, char *peer,
DDS_Long sleep_time, DDS_Long count)
{
    DDSSubscriber *subscriber = NULL;
    DDSDataReader *datareader = NULL;
    ShapeTypeExtendedReaderListener *listener  = new ShapeTypeExtendedReaderListener();
    DDS_DataReaderQos dr_qos;
    DDS_ReturnCode_t retcode;
    Application *application = NULL;
    RT_Registry_T *registry = NULL;
	
	

    application = new Application();
    if (application == NULL)
    {
        printf("failed Application new\n");
        goto done;
    }

	retcode = application->initialize("subscriber",
		"publisher",
		domain_id,
		udp_intf,
		peer,
		sleep_time,
		count);
	
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed Application initialize\n");
        goto done;
    }


	/////////////////////////////////
	// Initializate the subscriber //
	/////////////////////////////////
    subscriber = application->participant->create_subscriber(
        DDS_SUBSCRIBER_QOS_DEFAULT,NULL,
        DDS_STATUS_MASK_NONE);
    if (subscriber == NULL)
    {
        printf("subscriber == NULL\n");
        goto done;
    }

    retcode = subscriber->get_default_datareader_qos(dr_qos);
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed get_default_datareader_qos\n");
        goto done;
    }

    dr_qos.resource_limits.max_samples = 32;
    dr_qos.resource_limits.max_instances = 10;
    dr_qos.resource_limits.max_samples_per_instance = 32;
    /* if there are more remote writers, you need to increase these limits */
    dr_qos.reader_resource_limits.max_remote_writers = 10;
    dr_qos.reader_resource_limits.max_remote_writers_per_instance = 10;
    dr_qos.history.depth = 32;

    /* Reliability QoS */
    #if 1
    dr_qos.reliability.kind = DDS_RELIABLE_RELIABILITY_QOS;
    #else
    dr_qos.reliability.kind = DDS_BEST_EFFORT_RELIABILITY_QOS;
    #endif

    /* Only DDS_DATA_AVAILABLE_STATUS supported currently */
	datareader = subscriber->create_datareader(
        application->topic,
        dr_qos,
        listener,
        DDS_DATA_AVAILABLE_STATUS);
    
    if (datareader == NULL)
    {
        printf("datareader == NULL\n");
        goto done;
    }


	/////////////////////////////////
	// Initializate the publisher  //
	/////////////////////////////////
	

	printf("ENABLING APPLICATION\n");
    retcode = application->enable();
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed to enable application\n");
        goto done;
    }

	printf("Enabled\n");
    if (application->count != 0)
    {
        printf("Running for %d seconds, press Ctrl-C to exit\n",
        application->count);
        OSAPI_Thread_sleep(application->count * 1000);
    }
    else
    {
        int sleep_loop_count =  (24 * 60 * 60) / 2000;
        int sleep_loop_left = (24 * 60 * 60) % 2000;

        printf("Running for 24 hours, press Ctrl-C to exit\n");
		Sleep(30000);
        while (sleep_loop_count)
        {
            OSAPI_Thread_sleep(2000  * 1000); /* sleep is in ms */
            --sleep_loop_count;
        }

        OSAPI_Thread_sleep(sleep_loop_left * 1000);
    }

    done:

    if (subscriber != NULL)
    {
        if (datareader != NULL)
        {
            retcode = subscriber->delete_datareader(datareader);
            if (retcode != DDS_RETCODE_OK)
            {
                printf("failed to delete datareader\n");
                return -1;
            }
        }

        retcode = application->participant->delete_subscriber(subscriber);
        if (retcode != DDS_RETCODE_OK)
        {
            printf("failed to delete subscriber\n");
            return -1;
        }
    }

	

    if (application != NULL)
    {
        delete application;
    }

    if (listener != NULL)
    {
        delete listener;
    }
    DDS_DataReaderQos_finalize(&dr_qos);

    return 0;
}

#if !(defined(RTI_VXWORKS) && !defined(__RTP__))
int
main(int argc, char **argv)
{
    DDS_Long i = 0;
    DDS_Long domain_id = 0;
	DDS_Long publication_domain_id = -1;
    char *peer = NULL;
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;

    for (i = 1; i < argc; ++i)
    {
        if (!strcmp(argv[i], "-domain"))
        {
            ++i;
            if (i == argc)
            {
                printf("-domain <domain_id>\n");
                return -1;
            }
            domain_id = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-udp_intf"))
        {
            ++i;
            if (i == argc)
            {
                printf("-udp_intf <interface>\n");
                return -1;
            }
            udp_intf = argv[i];
        }
        else if (!strcmp(argv[i], "-peer"))
        {
            ++i;
            if (i == argc)
            {
                printf("-peer <address>\n");
                return -1;
            }
            peer = argv[i];
        }
        else if (!strcmp(argv[i], "-sleep"))
        {
            ++i;
            if (i == argc)
            {
                printf("-sleep_time <sleep_time>\n");
                return -1;
            }
            sleep_time = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-count"))
        {
            ++i;
            if (i == argc)
            {
                printf("-count <count>\n");
                return -1;
            }
            count = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-h"))
        {
            Application::help(argv[0]);
            return 0;
        }
		else if(!strcmp(argv[i], "-rejectColor"))
		{
			++i;
			if(i == argc)
			{
				printf("-rejectColor <Color_to_reject>");
				return -1;
			}
			sColor[nColors]= new char[20];
			strcpy(sColor[nColors], argv[i]);
			int aux_i = 0;
			while(sColor[nColors][aux_i]){
				sColor[nColors][aux_i] = toupper(sColor[nColors][aux_i]);
				aux_i++;
			}
			printf("Filter criteria: Color must not be: %s\n", sColor[nColors]);
			nColors++;
		}
		else if(!strcmp(argv[i], "-xMin")){
			++i;
			if(i == argc)
			{
				printf("-xMin <X minumum value>");
				return -1;
			}
			xMin = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Minimum x value: %d\n", xMin);
		}
		else if(!strcmp(argv[i], "-yMin")){
			++i;
			if(i == argc)
			{
				printf("-yMin <Y minumum value>");
				return -1;
			}
			yMin = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Minimum y value: %d\n", yMin);
		}
		else if(!strcmp(argv[i], "-xMax")){
			++i;
			if(i == argc)
			{
				printf("-xMax <X maximum value>");
				return -1;
			}
			xMax = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Maximum X value: %d\n", xMax);
		}
		else if(!strcmp(argv[i], "-yMax")){
			++i;
			if(i == argc)
			{
				printf("-yMax <Y maximum value>");
				return -1;
			}
			yMax = strtol(argv[i], NULL, 0);
			printf("Filter criteria: Maximum Y value: %d\n", yMax);
		}
		else if(!strcmp(argv[i], "-publish_domain")){
			++i;
			if(i == argc)
			{
				printf("-publish_domain <domain_id>");
				return -1;
			}
			publication_domain_id = strtol(argv[i], NULL, 0);
			printf("Enabled sample publication on domain: %d\n", publication_domain_id);
		}
        else
        {
            printf("unknown option: %s\n", argv[i]);
            return -1;
        }

    }

    return subscriber_main_w_args(domain_id, publication_domain_id, udp_intf, peer, sleep_time, count);
}
#elif defined(RTI_VXWORKS)
int
subscriber_main(void)
{
    /* Explicitly configure args below */
    DDS_Long domain_id = 44;
    char *peer = "10.10.65.103";
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;

    return subscriber_main_w_args(domain_id, udp_intf, peer, sleep_time, count);
}
#endif
